"""
This package contains a simple ActivityPub server for
testing purposes. See [Bovine - Tutorial-Server][https://bovine.readthedocs.io/en/latest/tutorials/server/].


Default variables are

```pycon
>>> hostname, port, handle_name
('bovine', 80, 'milkdrinker')

```

they can be set using `BOVINE_TEST_HOSTNAME`, `-_PORT`, `-_NAME`.

"""

import os
import secrets
import json

from quart import Quart, request

from bovine import BovineActor
from bovine.activitystreams import Actor
from bovine.crypto import build_validate_http_signature
from bovine.crypto.types import CryptographicIdentifier
from bovine.utils import webfinger_response_json

from .config import public_key, private_key


port = os.environ.get("BOVINE_TEST_PORT", 80)
hostname = os.environ.get("BOVINE_TEST_HOSTNAME", "bovine")
handle_name = os.environ.get("BOVINE_TEST_NAME", "milkdrinker")

actor_id = f"http://{hostname}/{handle_name}"
"""The actor id"""

webfinger_response = webfinger_response_json(f"acct:{handle_name}@{hostname}", actor_id)
"""The webfinger response"""

actor_object = Actor(
    id=actor_id,
    preferred_username=handle_name,
    name="The Milk Drinker",
    inbox=f"http://{hostname}/inbox",
    outbox=actor_id,
    public_key=public_key,
    public_key_name="main-key",
).build()


actor = BovineActor(
    actor_id=actor_id,
    public_key_url=f"{actor_id}#main-key",
    secret=private_key,
)


def make_id():
    return f"http://{hostname}/" + secrets.token_urlsafe(6)


def create_app():
    app = Quart(__name__)

    @app.before_serving
    async def startup():
        await actor.init()

    async def fetch_public_key(url):
        result = await actor.get(url)
        return CryptographicIdentifier.from_public_key(result["publicKey"])

    verify = build_validate_http_signature(fetch_public_key)

    @app.get("/.well-known/webfinger")
    async def webfinger():
        return webfinger_response

    @app.get("/" + handle_name)
    async def get_actor():
        return actor_object, 200, {"content-type": "application/activity+json"}

    @app.post("/inbox")
    async def post_inbox():
        controller = await verify(request)
        if not controller:
            return "ERROR", 401

        data = await request.get_json()
        print(f"Received in inbox from {controller}")
        print(json.dumps(data, indent=2))
        print()
        return "success", 202

    return app



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

"""

<!--
```
>>> bovine.utils.now_isoformat = lambda: '2024-09-26T18:35:42Z'

```
-->
"""

from dataclasses import dataclass, field
from typing import Optional, Set, Callable

import bovine.utils
from .utils import id_for_object


@dataclass
class Activity:
    """A dataclass representing an [ActivityStreams Activity](https://www.w3.org/TR/activitystreams-vocabulary/#activity-types).

    ```pycon
    >>> activity=Activity(
    ...  type="Like",
    ...  actor="http://actor.example",
    ... object="http://some.object.example")
    >>> activity.build()
    {'@context': 'https://www.w3.org/ns/activitystreams',
        'type': 'Like',
        'actor': 'http://actor.example',
        'object': 'http://some.object.example'}

    ```

    """

    type: str
    actor: Optional[str] = None
    followers: Optional[str] = None
    id: Optional[str] = None
    published: Optional[str] = None
    to: Set[str] = field(default_factory=set)
    cc: Set[str] = field(default_factory=set)

    name: Optional[str] = None
    summary: Optional[str] = None
    content: Optional[str] = None

    target: Optional[str] = None
    object: Optional[str] = None

    def as_public(self):
        """makes the activity public, i.e. public in to and followers in cc

        ```pycon
        >>> activity = Activity(type="Like",
        ...     actor="http://actor.example",
        ...     followers="http://actor.example/followers")
        >>> activity.as_public().build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Like',
            'actor': 'http://actor.example',
            'to': ['https://www.w3.org/ns/activitystreams#Public'],
            'cc': ['http://actor.example/followers']}

        ```
        """
        self.to.add("https://www.w3.org/ns/activitystreams#Public")
        if self.followers:
            self.cc.add(self.followers)
        return self

    def as_followers(self):
        """addresses the activity to followers, if they are set

        ```pycon
        >>> activity = Activity(type="Like",
        ...     actor="http://actor.example",
        ...     followers="http://actor.example/followers")
        >>> activity.as_followers().build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Like',
            'actor': 'http://actor.example',
            'to': ['http://actor.example/followers']}

        ```
        """
        if self.followers:
            self.to.add(self.followers)
        return self

    def as_unlisted(self):
        """makes the activity unlisted, i.e. public in cc and followers in to

        ```pycon
        >>> activity = Activity(type="Like",
        ...     actor="http://actor.example",
        ...     followers="http://actor.example/followers")
        >>> activity.as_unlisted().build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Like',
            'actor': 'http://actor.example',
            'to': ['http://actor.example/followers'],
            'cc': ['https://www.w3.org/ns/activitystreams#Public']}

        ```
        """
        if self.followers:
            self.to.add(self.followers)
        self.cc.add("https://www.w3.org/ns/activitystreams#Public")
        return self

    def build(self) -> dict:
        """converts the activity into a dict, that can be serialized to JSON"""
        result = {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": self.type,
            "actor": self.actor,
            "to": list(self.to),
            "cc": list(self.cc - self.to),
        }

        extra_fields = {
            "id": self.id,
            "published": self.published,
            "name": self.name,
            "summary": self.summary,
            "content": self.content,
            "target": self.target,
            "object": self.object,
        }

        if result["to"] is not None and len(result["to"]) == 0:
            del result["to"]
        if result["cc"] is not None and len(result["cc"]) == 0:
            del result["cc"]

        for key, value in extra_fields.items():
            if value:
                result[key] = value

        return result


def recipient(value):
    if isinstance(value, str):
        return {value}
    return set(value)


@dataclass
class ActivityFactory:
    """Basic factory for Activity objects.
    Can created by [BovineClient.activity_factory][bovine.BovineClient.activity_factory]

    ```pycon
    >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
    >>> activity_factory.like("http://object.example").build()
    {'@context': 'https://www.w3.org/ns/activitystreams',
        'type': 'Like',
        'actor': 'http://actor.example',
        'published': '2024-09-26T18:35:42Z',
        'object': 'http://object.example'}

    ```

    By setting id_generator, one can provide a function that will automatically set the id property:

    ```pycon
    >>> activity_factory = ActivityFactory({"id": "http://actor.example"},
    ...     id_generator=lambda: "http://actor.example/id")
    >>> activity_factory.like("http://object.example").build()
    {'@context': 'https://www.w3.org/ns/activitystreams',
        'type': 'Like',
        'actor': 'http://actor.example',
        'id': 'http://actor.example/id',
        'published': '2024-09-26T18:35:42Z',
        'object': 'http://object.example'}

    ```
    """

    actor_information: dict
    id_generator: Callable[[], str] | None = None

    def _defaults_for_object(self, obj, kwargs):
        result = {
            "actor": self.actor_information.get("id", obj.get("attributedTo")),
            "object": obj,
            "cc": recipient(obj.get("cc", [])),
            "to": recipient(obj.get("to", [])),
            "published": bovine.utils.now_isoformat(),
            **kwargs,
        }
        if self.id_generator:
            result["id"] = self.id_generator()
        return result

    def create(self, obj, **kwargs):
        """Activity of type Create from Object

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> obj = {"type": "Note", "content": "hello world!", "to": "http://you.example"}
        >>> activity_factory.create(obj).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Create',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': {'type': 'Note',
                'content': 'hello world!',
                'to': 'http://you.example'}}

        ```
        """
        return Activity(
            type="Create",
            **self._defaults_for_object(obj, kwargs),
        )

    def update(self, obj, **kwargs):
        """Activity of type Update from Object"""
        return Activity(
            type="Update",
            **self._defaults_for_object(obj, kwargs),
        )

    def _base_defaults(self, kwargs):
        result = {
            "actor": self.actor_information["id"],
            "published": bovine.utils.now_isoformat(),
            **kwargs,
        }
        if self.actor_information.get("followers") and not result.get("followers"):
            result["followers"] = self.actor_information.get("followers")

        if self.id_generator:
            result["id"] = self.id_generator()

        return result

    def _defaults(self, target, kwargs):
        return {"object": target, **self._base_defaults(kwargs)}

    def like(self, target, **kwargs):
        """Like for target

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> activity_factory.like("http://object.example").build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Like',
            'actor': 'http://actor.example',
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://object.example'}

        ```


        """
        return Activity(type="Like", **self._defaults(target, kwargs))

    def delete(self, target, **kwargs):
        """Delete for target

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> activity_factory.delete("http://bad.example", to={"http://you.example"}).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Delete',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://bad.example'}

        ```
        """
        return Activity(type="Delete", **self._defaults(target, kwargs))

    def accept(self, activity, include_activity=False, **kwargs):
        """Accept for object

        ```pycon
        >>> follow = ActivityFactory({"id":
        ...     "http://you.example"}).follow("http://actor.example",
        ...     id="http://actor.example/follow_id").build()
        >>> follow
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Follow',
            'actor': 'http://you.example',
            'to': ['http://actor.example'],
            'id': 'http://actor.example/follow_id',
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://actor.example'}

        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> activity_factory.accept(follow).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Accept',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://actor.example/follow_id'}

        ```
        """

        if isinstance(activity, str):
            return Activity(type="Accept", **self._defaults(activity, kwargs))

        obj = id_for_object(activity)
        if obj is None or include_activity:
            obj = activity

        return Activity(
            type="Accept",
            **self._defaults(obj, kwargs),
            to={id_for_object(activity.get("actor"))},
        )

    def reject(self, activity, include_activity=False, **kwargs):
        """Reject for object

        ```pycon
        >>> follow = ActivityFactory({"id":
        ...     "http://you.example"}).follow("http://actor.example",
        ...     id="http://actor.example/follow_id").build()
        >>> follow
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Follow',
            'actor': 'http://you.example',
            'to': ['http://actor.example'],
            'id': 'http://actor.example/follow_id',
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://actor.example'}

        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> activity_factory.reject(follow).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Reject',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://actor.example/follow_id'}

        ```
        """

        if isinstance(activity, str):
            return Activity(type="Reject", **self._defaults(activity, kwargs))

        obj = id_for_object(activity)
        if obj is None or include_activity:
            obj = activity

        return Activity(
            type="Reject",
            **self._defaults(obj, kwargs),
            to={id_for_object(activity.get("actor"))},
        )

    def announce(self, obj, **kwargs):
        """Announce for object

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example",
        ...     "followers": "http://actor.example/followers"})
        >>> activity_factory.announce("http://object.example").as_public().build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Announce',
            'actor': 'http://actor.example',
            'to': ['https://www.w3.org/ns/activitystreams#Public'],
            'cc': ['http://actor.example/followers'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://object.example'}

        ```
        """
        return Activity(
            type="Announce",
            **self._defaults(obj, kwargs),
        )

    def follow(self, obj: str | dict, **kwargs):
        """Follow for object

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> activity_factory.follow("http://you.example").build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Follow',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://you.example'}

        ```

        If the object is an actor, its id is used

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> actor = {"type": "Person", "id": "http://you.example"}
        >>> activity_factory.follow(actor).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Follow',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://you.example'}

        ```


        :param obj: Object to be followed
        :param **kwargs: Passed to [Activity][bovine.activitystreams.activity_factory.Activity]'s constructor

        """

        obj = id_for_object(obj)

        return Activity(
            type="Follow",
            **self._defaults(obj, kwargs),
            to={obj},
        )

    def undo(self, activity, include_activity=False, **kwargs):
        """Undo for activity

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> follow = activity_factory.follow("http://you.example",
        ...     id="http://actor.example/follow_id").build()
        >>> activity_factory.undo(follow).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Undo',
            'actor': 'http://actor.example',
            'to': ['http://you.example'],
            'published': '2024-09-26T18:35:42Z',
            'object': 'http://actor.example/follow_id'}

        ```

        If the activity doesn't have an id, it is fully included.


        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example",
        ...     "followers": "http://actor.example/followers"})
        >>> announce = activity_factory.announce("http://object.example").as_public().build()
        >>> activity_factory.undo(announce).build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'Undo',
            'actor': 'http://actor.example',
            'to': ['https://www.w3.org/ns/activitystreams#Public'],
            'cc': ['http://actor.example/followers'],
            'published': '2024-09-26T18:35:42Z',
            'object': {'@context': 'https://www.w3.org/ns/activitystreams',
                'type': 'Announce',
                'actor': 'http://actor.example',
                'to': ['https://www.w3.org/ns/activitystreams#Public'],
                'cc': ['http://actor.example/followers'],
                'published': '2024-09-26T18:35:42Z',
                'object': 'http://object.example'}}


        ```
        """

        obj = id_for_object(activity)
        if obj is None or include_activity:
            obj = activity

        return Activity(
            type="Undo",
            **self._defaults(obj, kwargs),
            to={id_for_object(x) for x in activity.get("to", [])},
            cc={id_for_object(x) for x in activity.get("cc", [])},
        )

    def custom(self, **kwargs):
        """Allows creating a custom activity

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example"})
        >>> activity_factory.custom(type="AnimalSound", content="moo").build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'AnimalSound',
            'actor': 'http://actor.example',
            'published': '2024-09-26T18:35:42Z',
            'content': 'moo'}

        ```

        Or ready to send to your followers

        ```pycon
        >>> activity_factory = ActivityFactory({"id": "http://actor.example",
        ...     "followers": "http://actor.example/followers"})
        >>> activity_factory.custom(type="AnimalSound", content="moo").as_public().build()
        {'@context': 'https://www.w3.org/ns/activitystreams',
            'type': 'AnimalSound',
            'actor': 'http://actor.example',
            'to': ['https://www.w3.org/ns/activitystreams#Public'],
            'cc': ['http://actor.example/followers'],
            'published': '2024-09-26T18:35:42Z',
            'content': 'moo'}

        ```


        """

        return Activity(**self._base_defaults(kwargs))



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from typing import List


@dataclass
class Multikey:
    """Represents a Multikey"""

    id: str
    controller: str
    multibase: str
    type: str = "Multikey"

    def build(self, include_context=True):
        result = {
            "id": self.id,
            "type": self.type,
            "controller": self.controller,
            "publicKeyMultibase": self.multibase,
        }
        if include_context:
            result["@context"] = "https://w3id.org/security/multikey/v1"
        return result

    @staticmethod
    def from_multibase_and_controller(controller, multibase):
        return Multikey(
            id=f"{controller}#{multibase}", controller=controller, multibase=multibase
        )


@dataclass
class Controller:
    """Experimental class to represent a controller document
    see [FEP-521a](https://codeberg.org/fediverse/fep/src/branch/main/fep/521a/fep-521a.md).

    See also the recent W3C Draft: [Controller Documents](https://www.w3.org/TR/controller-document/)

    ```pycon
    >>> multikey = Multikey(
    ... id="https://server.example/users/alice#ed25519-key",
    ... controller="https://server.example/users/alice",
    ... multibase="z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2")
    >>> Controller(assertion_method=[multikey]).build()
    {'@context': ['https://www.w3.org/ns/did/v1',
        'https://w3id.org/security/multikey/v1'],
        'assertionMethod':
            [{'id': 'https://server.example/users/alice#ed25519-key',
            'type': 'Multikey',
            'controller': 'https://server.example/users/alice',
            'publicKeyMultibase': 'z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2'}]}

    ```
    """

    assertion_method: List[Multikey] = field(default_factory=list)
    authentication: List[Multikey] = field(default_factory=list)

    def build(self):
        """Creates the controller document. Currently only assertion_method is supported"""
        if len(self.assertion_method) == 0:
            return {}

        return {
            "@context": [
                "https://www.w3.org/ns/did/v1",
                "https://w3id.org/security/multikey/v1",
            ],
            "assertionMethod": [
                key.build(include_context=False) for key in self.assertion_method
            ],
        }



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

"""Helper classes to parse signatures from the HTTP Headers"""

from dataclasses import dataclass
import logging
import http_sf
from typing import List


logger = logging.getLogger(__name__)


def int_or_none(value):
    if value:
        return int(value)
    return None


@dataclass
class Signature:
    """Helper class to parse HTTP Signatures"""

    key_id: str
    """key_id"""
    algorithm: str
    """The algorithm used"""
    headers: str
    """The header fields contained in the signature"""
    signature: str
    """The signature string"""
    created: int | None = None
    """Created timestamp"""
    expires: int | None = None
    """The expires timestamp"""

    def __post_init__(self):
        if self.algorithm not in ["rsa-sha256", "hs2019"]:
            logger.error(f"Unsupported algorithm {self.algorithm}")
            logger.error(self.signature)
            logger.error(self.headers)
            logger.error(self.key_id)

            raise ValueError("Unsupported algorithm", self.algorithm)

    @property
    def fields(self) -> List[str]:
        """Returns the fields that are used when building the signature"""
        return self.headers.split(" ")

    @staticmethod
    def from_signature_header(header):
        """Takes the signature header and turns into Signature object

        The header is assumed of the for key=value,... The keys keyId,
        algorithm, headers, and signature are parsed. If algorithm
        is absent it is assumed to be rsa-sha256. The other keys are required.

        ```pycon
        >>> header = 'keyId="https://host.user#main-key",algorithm="rsa-sha256",headers="(request-target) host date digest content-type",signature="h...Kg=="'
        >>> Signature.from_signature_header(header)
        Signature(key_id='https://host.user#main-key',
            algorithm='rsa-sha256',
            headers='(request-target) host date digest content-type',
            signature='h...Kg==',
            created=None,
            expires=None)

        ```

        (created) and (expires) are supported via

        ```pycon
        >>> header = 'keyId="https://key.example",algorithm="hs2019",headers="host date (request-target) (created) (expires)",signature="s0...",created="1728671105",expires="1728674705"'
        >>> Signature.from_signature_header(header)
        Signature(key_id='https://key.example',
            algorithm='hs2019',
            headers='host date (request-target) (created) (expires)',
            signature='s0...',
            created=1728671105,
            expires=1728674705)

        ```
        """
        headers = header.split(",")
        headers = [x.split('="', 1) for x in headers]
        parsed = {x[0]: x[1].replace('"', "") for x in headers}

        created = int_or_none(parsed.get("created", None))
        expires = int_or_none(parsed.get("expires", None))
        return Signature(
            parsed["keyId"],
            parsed.get("algorithm", "rsa-sha256"),
            parsed["headers"],
            parsed["signature"],
            created=created,
            expires=expires,
        )


@dataclass
class RFC9421Signature:
    """Helper class to parse signatures according to
    [RFC 9421](https://www.rfc-editor.org/rfc/rfc9421.html).

    Currently, RFC9421 support is in development in bovine"""

    fields: List[str]
    """fields used in the signature"""
    signature: bytes
    """The signature"""
    signature_params: str
    """The signature parameters as a string"""
    params: dict
    """The signature parameters as a dictionary"""

    @property
    def key_id(self):
        """
        Returns the id of the public key

        ```pycon
        >>> signature = RFC9421Signature(fields=[],
        ...     signature=b"",
        ...     params={"keyid": "http://actor.example/key"},
        ...     signature_params=b"")
        >>> signature.key_id
        'http://actor.example/key'

        ```
        """

        return self.params["keyid"]

    @staticmethod
    def from_headers(signature_input, signature):
        """
        ```pycon
        >>> signature_input='sig-b26=("date" "@method" "@path" "@authority" "content-type" "content-length");created=1618884473;keyid="test-key-ed25519"'
        >>> signature='sig-b26=:wqcAqbmYJ2ji2glfAMaRy4gruYYnx2nEFN2HN6jrnDnQCK1u02Gb04v9EDgwUPiu4A0w6vuQv5lIp5WPpBKRCw==:'
        >>> RFC9421Signature.from_headers(signature_input, signature)
        RFC9421Signature(fields=['date', '@method', '@path', '@authority', 'content-type', 'content-length'],
            signature=b"...",
            signature_params='("date" "@method" "@path" "@authority" "content-type" "content-length");created=1618884473;keyid="test-key-ed25519"',
            params={'created': 1618884473, 'keyid': 'test-key-ed25519'})

        ```
        """

        parsed_input = http_sf.parse_dictionary(signature_input.encode())[1]
        parsed_signature = http_sf.parse_dictionary(signature.encode())[1]

        if len(parsed_input) != 1:
            raise ValueError("Currently only one signature is supported")

        signature_key = list(parsed_input.keys())[0]
        parsed_input = parsed_input[signature_key]
        parsed_signature = parsed_signature[signature_key]

        fields = [x[0] for x in parsed_input[0]]

        return RFC9421Signature(
            fields=fields,
            signature=parsed_signature[0],
            params=parsed_input[1],
            signature_params=signature_input.split("=", 1)[1],
        )



import asyncio
import aiohttp
import logging

import click


from bovine.activitystreams import factories_for_actor_object
from bovine.clients import lookup_uri_with_webfinger

from ptpython.repl import embed

from .server import create_app, actor, actor_object, make_id

activity_factory, object_factory = factories_for_actor_object(
    actor_object, id_generator=make_id
)


@click.group()
@click.option(
    "--debug", is_flag=True, default=False, help="Sets the log level to debug."
)
def main(debug):
    """Command line to tool to manage a testing environment"""

    logging.captureWarnings(True)

    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)


@main.command()
@click.option("--port", default=5000, help="Port the application runs on")
@click.option("--reload", default=False, is_flag=True, help="Enable auto reloading")
def serve(port, reload):
    """Serves the app from the bovine tutorial as
    a server"""
    app = create_app()
    app.run(port=port, host="0.0.0.0", use_reloader=reload)


def config(repl):
    repl.use_code_colorscheme("dracula")
    repl.enable_output_formatting = True


async def repl():
    print(
        "Relevant variables are actor, actor_object, activity_factory, object_factory."
    )

    async with aiohttp.ClientSession() as session:
        await actor.init(session=session)

        async def webfinger(acct_uri):
            domain = acct_uri.split("@")[1]
            actor_id, _ = await lookup_uri_with_webfinger(
                session, acct_uri, domain=f"http://{domain}"
            )
            return actor_id

        await embed(
            globals=globals(),
            locals=locals(),
            return_asyncio_coroutine=True,
            patch_stdout=True,
            configure=config,
        )


@main.command()
def shell():
    """Opens a REPL to perform actions using the [BovineActor][bovine.BovineActor]
    from the server tutorial"""
    asyncio.run(repl())


@main.command()
@click.option("--nano", help="Use nano editor", is_flag=True, default=False)
def edit(nano):
    """Allows one to edit the main server file"""

    editor = "nano" if nano else "vim"

    import subprocess

    subprocess.run(["apk", "add", editor])
    import bovine.testing.server

    subprocess.run([editor, bovine.testing.server.__file__])


if __name__ == "__main__":
    main()
